from intelli_utilities import fetch_hist_data, raise_bai_exception, get_closest_index, get_bin_range, LOCAL_TIME_ZONE
import pandas as pd
from datetime import datetime

class DataProcessor():
    
    def __init__(self, token:str, base_url:str):
        self._token = token
        self._base_url = base_url
        self._EXPLANATORY_VAR_PREFIX = 'explanatory_var_'                  
        
    def fetch_and_process_data(self, node:dict, interval:int, fetch_dates: dict[str, str]):
        hist_data: list = []
        hist_data = self.get_hist_data_from_dates(node, node['interpolation'], fetch_dates, interval)

        if len(hist_data) == 0:
            raise_bai_exception('mainNodeNoData')
        data_df = self.convert_histdata_to_dataframe(hist_data)

        return data_df
    
    def get_hist_data_from_dates(self, pmModel:dict, interpolation:dict, interval:int, dates: dict[str, str]) -> list:
        histdata = []
        for date in dates:
            histdata = histdata + fetch_hist_data(pmModel, interpolation, interval, self._token, date['start'], date['end'])

        if len(histdata) == 0:
            raise_bai_exception('noDataFound', {'path': pmModel['path']})

        return histdata
    
    def convert_histdata_to_dataframe(self, dms_hist_data:list, value_col_name = 'value'):
        """ 
        @param hist_data: list
        Converts dms format: [{'2023-03-26T01:00:00,000+01:00': 23}, {'2023-03-26T03:00:00,000+02:00': 44}]
        to a dataframe format: {'x': ['2023-03-26 01:00:00+01:00', '2023-03-26 03:00:00+02:00'], 'y': [23,44]}

        Method "pd.to_datetime" fails converting data with mixed offset to DateTimeIndex. More: https://pandas.pydata.org/docs/reference/api/pandas.to_datetime.html#to-datetime-tz-examples
        Converting each date individually works.
        """
        keys = []
        values = []
        for el in dms_hist_data:
            x = list(el.keys())[0]
            x = datetime.fromisoformat(x)
            x = pd.to_datetime(x).tz_convert(LOCAL_TIME_ZONE)
            keys.append(x)
            values.append(list(el.values())[0])

        df = pd.DataFrame({
            '{}'.format(value_col_name): values
        }, index=keys)
        
        if df.empty:
            raise_bai_exception('No data found for column "{}"'.format(value_col_name))
            
        return df    
        
    def binarize_data(self, data_df: pd.DataFrame, col_name:str, bins:dict) -> pd.DataFrame:
        """ 
        @param data_df: df to update
        @param features: List of features to binarize. ie. [{'col_name': {bins: {'base': 0, 'step': 5}}}]

        """
        highest_value = max(data_df[col_name].values)
        lowest_value = min(data_df[col_name].values)    
        bin_range = get_bin_range(bins['base'], bins['step'], lowest_value, highest_value)
        data_df[col_name + '_binned'] = data_df[col_name].apply(get_closest_index, bin_range=bin_range)
    
        return data_df            
    
    # add possible group-by variables
    def add_temporal_variables(self, data:pd.DataFrame):
        #minute of hour
        data["moh"] = data.index.minute
        #hour of day
        data["hod"] = data.index.hour
        # day of week
        data["dow"] = data.index.dayofweek
        # week
        data["woy"] = data.index.isocalendar().week
        # Month
        data["month"] = data.index.month 
        # quarter
        data["quarter"] = data.index.quarter     
        # distinction between Monday to Saturday vs Sunday
        sunday = data.index.dayofweek == 6
        # data["Mo2Sa-So"] = 'Mo2Sa'
        # data.loc[sunday, "Mo2Sa-So"] = 'So'
        data["Mo2Sa-So"] = 1
        data.loc[sunday, "Mo2Sa-So"] = 2
        # distinction between Monday to Friday vs Sa vs Su
        saturday = data.index.dayofweek == 5
        data["Mo2Fr-Sa-So"] = 'Mo2Fr'
        # data.loc[saturday, "Mo2Fr-Sa-So"] = 'Sa'
        # data.loc[sunday, "Mo2Fr-Sa-So"] = 'So'
        data["Mo2Fr-Sa-So"] = 1
        data.loc[saturday, "Mo2Fr-Sa-So"] = 2
        data.loc[sunday, "Mo2Fr-Sa-So"] = 3    
        # distinction between Monday to Friday vs Weekend
        data["Mo2Fr-Sa2So"] = 1
        data.loc[saturday, "Mo2Fr-Sa2So"] = 2
        data.loc[sunday, "Mo2Fr-Sa2So"] = 2
        # data.loc[saturday, "Mo2Fr-Sa2So"] = 'Sa2So'
        # data.loc[sunday, "Mo2Fr-Sa2So"] = 'Sa2So'
        return data 