from baseClass import BaseThresholdClass
import pandas as pd
import numpy as np

class QuantileThresholds(BaseThresholdClass):
    """This class calculated the threholds based on Quantiles per group.

    Thresholds are calculated by fit-method)
    1. grouping data into different folds (accoring to 'group_by') and
    2. determining the quantiles per _individual_ group.
    The calculated thresholds are applied (transform-method) to the identically grouped evaluation data.
    """
    def __init__(self, group_by=['dow', 'range_Tout'],
                 q_lower=0.01, q_upper=0.99,
                 min_no_elements_per_bin = 5, **kwargs):

        self._group_by = group_by
        self._q_lower = q_lower
        self._q_upper = q_upper
        self._min_no_elements_per_bin = min_no_elements_per_bin
        self.model_info = None
        self.thresholds_df = None

    def fit(self, data:pd.DataFrame) -> dict:
        """Create thresholds for each bin based on provided data"""
        # TODO: idealy add checks that columns in self._group_by are available in data
        # Calculate thresholds for all combinations of "day of week" and "T_out bins"
        grouped = data.groupby(by=self._group_by)
        # get thresholds
        th_lower = grouped.quantile(q=self._q_lower)[['y']].rename(columns={'y':'lower'}) # type:pd.DataFrame
        th_upper = grouped.quantile(q=self._q_upper)[['y']].rename(columns={'y':'upper'}) # type:pd.DataFrame
        thresholds_df = th_lower.join(th_upper)
        del th_upper, th_lower
        # get size of each group
        counts_df = pd.DataFrame(grouped.size(), columns=['size'])
        self.thresholds_df = thresholds_df.join(counts_df)
        self.model_info={
            'params': None,
            'info'  : self.thresholds_df['size'].to_dict(),
             }
        return None

    def _get_range(self, data_row:pd.Series):
        """get acceptable range for the corresponding tuple."""
        idx_tuple = tuple(data_row[self._group_by].values)
        # handle case where group_by contains only single element
        idx_tuple = idx_tuple[0] if len(self._group_by) == 1 else idx_tuple
        if idx_tuple in self.thresholds_df.index:
            if self.thresholds_df.loc[idx_tuple, "size"] <= self._min_no_elements_per_bin:
                # in case the number of elements is smaller than _min_no_elements_per_bin
                return [np.NINF, np.inf, self.thresholds_df.loc[idx_tuple, "size"]]
            else:
                return self.thresholds_df.loc[idx_tuple, :]
        else:
            # in case there was no bin for this combination in the training data
            return [np.NINF, np.inf, 0.0]

    def transform(self, data:pd.DataFrame) -> pd.DataFrame:
        """Return data with column added for (i) thresholds and (ii) distance to closeses threshold

        The distance is only provided if the value is outside of the thresholds"""
        # check that self.thresholds_df is populated
        if self.thresholds_df is None:
            raise ValueError("Call 'fit' before 'transform'")
        # get thesholds
        data = data.join(data.apply(self._get_range, axis=1, result_type='expand'))
        data = data.rename(columns={0 : 'lower', 1 : 'upper', 2 : 'size'})
        # get minimal distance to threshold
        data = self._get_min_dist(data)
        return data