import pandas as pd
from typing import Union
import numpy as np
import os, sys

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(os.path.dirname(currentdir))
sys.path.append(parentdir)
from intelli_utilities import request_apigateway

# Abstract base Class defining the interface for the Threhold Calsses
class BaseThresholdClass():
    _decimals = 3

    def fit(self, data:pd.DataFrame) -> Union[dict, None]:
        """Create model based on the provided data
        
        :param pd.DataFrame data:  training data. Assumes that the column 'value'
            and the remaining columns correspond to the output (y) and input (X) 
            values of a classical ML problem, respectively
        :return: optionally returns a dict with keys 'params' and 'info', where 
                 'params' contains the model parameters as a dict and 
                 'info' contains additional information as a dict """
        pass

    def transform(self, data:pd.DataFrame) -> pd.DataFrame:
        """Predict the thresholds based on the model learned with the .fit method
        
        :param pd.DataFrame data:  Data for threshold prediction. Must have the same
            columns available as used in the .fit-method. (Potentially all) columns 
            beside the 'value' column are used to predict an estimate 'Y_pred'
            for the 'value'. Subsequently, lower and upper thresholds are derived. 
        :return pd.DataFrame: Return data with (at least) the following added columns: 
            * 'Y_pred'    : estimate of the value based on the learned model
            * 'lower'     : lower threshold
            * 'upper'     : upper threshold
            * 'min_dist'  : minimal distance of the 'value' to one of the thresholds or 
                            0.0 if the value lies between thresholds
            If the class has been initialized with a parameter 'explanatory_vars', only
            those columns (and the columns 'value') will be available in the returned 
            dataframe as to decrease the handing around of irrelevant data.
        """
        pass

    def _get_min_dist(self, data:pd.DataFrame) -> pd.DataFrame:
        """helper function to calculate the 'min_dist'

        :param data: DataFrame needs to have the columns ['value', 'upper', 'lower']
        """
        # get minimal distance to threshold
        data['dist_upper'] = np.round(np.abs(data[self.dependent_variable] - data['upper']), self._decimals)
        data['dist_lower'] = np.round(np.abs(data[self.dependent_variable] - data['lower']), self._decimals)

        index_zero = (data[self.dependent_variable] <= data['upper'])
        data.loc[index_zero, 'dist_upper'] = 0.0

        index_zero = (data[self.dependent_variable] >= data['lower'])
        data.loc[index_zero, 'dist_lower'] = 0.0
        return data

    def _is_sufficient_data_available(self, data:pd.DataFrame) -> bool:
        """Check if 80% of data has a valid value for explanatory_variable 
        """
        # investigate selected rows for nan values
        rows_without_nan_values = ~data.isna().any(axis=1) 
        return np.sum( rows_without_nan_values )/data.shape[0] >= 0.8 

    def select_binarized_data(self, explanatory_vars:list) -> list:
        # Select binarized data                
        temporal_vars = [
            'woy',
            'month',
            'quarter',
            'Mo2Sa-So',
            'Mo2Fr-Sa-So',
            'Mo2Fr-Sa2So',
            'hod',
            'dow',
        ]
        
        for i, var in enumerate(explanatory_vars):
            if var not in temporal_vars and var.startswith('holiday') == False:
                explanatory_vars[i] = var + '_binned'        
        
        return explanatory_vars

    def load_model(self, *args, **kwargs):
        """Define how to load the model"""
        pass

    def save_model(self, dir_name:str, headers:dict, process_manager:str):
        """Define how to save the model
        :param dir_name: save location. Usually yo want to set config_model_uuid
        :param token
        :param process_manager
        """             
        pass

    def request_apigateway(self, uri:str, process_manager:str, headers:dict = {}, method:str = 'get'):
        """Requests apigateway
        :param uri: uri to request
        :param process_manager
        :param headers: request headers
        :param method: request method
        """          
        return request_apigateway(uri, process_manager, headers, method)    
    
