from sklearn.preprocessing import StandardScaler 
import pandas as pd
import numpy as np


def calculate_metrics(results_df:pd.DataFrame, dependent_variable:str, punish_missing_values=100):
    results_df = results_df.copy()
    """Calculate events, alarms, metrics 
    :param pd.DataFrame results_df:  DataFrame containing the results from the function do_backtesting
    :return tuple: (pd.DataFrame, dict) Where:
        1. Return results_df with xxxx added columns originating from the 'thresholding class' during backtesting
            * 'event'      : True if the min_dist > 0.0, i.e. if the value is outside of the thresholds
            * 'alarm'      : True, see function `get_alarms` 
            * 'error'      : Distance between predicted (`Y_pred`) and measured value (`value`) 
            * 'errorE'     : "Extended Error" -> error but all positions without an error are filled with max(error)
            * 'SqrError'   : squared error
            * 'SqrErrorE'  : squared errorE
            * 'rollingMSE' : moving mean over sqrError with window length `lookback_horizon`
            * 'rollingMSEE': moving mean over sqrErrorE with window length `lookback_horizon`
            * 'rollingEventCount' : number events summed up over `lookback_horizon` 
        2. dict of summary statistics
    """
   
    # (squared) error 
    results_df['error'] = results_df[dependent_variable] - results_df['Y_pred']
    results_df['SqrError'] = np.square(results_df['error'])
    
    # "error Extended" -> squared error also for positions without Threshold
    max_abs_error = results_df['error'].abs().max()
    # Added punishment for na values as well
    idx_nothreshold = results_df[(results_df['upper'] == np.inf) | (results_df['upper'].isna())].index

    results_df['errorE'] = results_df['error'].copy()
    
    # Weighting the missing thresholds with punish_missing_values times the max. error of the the
    results_df.loc[idx_nothreshold, 'errorE'] =  punish_missing_values * max_abs_error 
    results_df['SqrErrorE'] = np.square(results_df['errorE'])
    
    # "Normalized error Extended"  
    stdScaler = StandardScaler().fit(results_df[dependent_variable].values.reshape(-1, 1))
    scaled_value = stdScaler.transform(results_df[dependent_variable].values.reshape(-1, 1))
    scaled_prediction = stdScaler.transform(results_df['Y_pred'].values.reshape(-1, 1))   
    
    # results_df['valueScaled'] = scaled_value
    # results_df['Y_predScaled'] = scaled_prediction
    results_df['errorScaled'] = scaled_value - scaled_prediction 
    # results_df['SqrErrorScaled'] = np.square(results_df['errorScaled'])
    
    # "error Extended" -> squared error also for positions without Threshold
    mean_abs_error_scaled = results_df['errorScaled'].abs().max()
    results_df['errorEScaled'] = results_df['errorScaled'].copy()
    results_df.loc[idx_nothreshold, 'errorEScaled'] = punish_missing_values * mean_abs_error_scaled
    results_df['SqrErrorEScaled'] = np.square(results_df['errorEScaled'])  
    
    mseEScaled = results_df['SqrErrorEScaled'].mean()
    
    return mseEScaled