import pandas as pd
import numpy as np
import os, sys
import uuid
import traceback
import argparse
import math
import time
import warnings
warnings.filterwarnings("ignore")

# Append maxpermissible values folder
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# Append scripts folder
parentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)

from lib.Thresholds import Thresholds, METHODS_CLASS_DICT
import json
from intelli_utilities import BaiException, LOCAL_TIME_ZONE
from metrics.calculate_metrics import calculate_metrics
import multiprocessing
from collections import deque 
from metrics.metric_utils import *
from datetime import datetime, timedelta
from websockets.sync.client import connect
from abc import ABC, abstractmethod

import logging
# logging.basicConfig(filename='log/jobs.log', encoding='utf-8', level=logging.ERROR)
logging.basicConfig(level=logging.ERROR, format='%(message)s', stream=sys.stderr)

import asyncio
           
# TODO: Wrap Sender implementation in own JobSender pip package
class Sender(ABC):
    @abstractmethod
    def send(self, progress: dict):
        pass

class LocalSender(Sender):
    def send(self, data: dict):
        print(json.dumps(data), flush=True)
    def error(self, data:str):
        logging.error(json.dumps(data))

class WebSocketSender(Sender):
    def __init__(self, ws, job_uuid:str):
        self.ws = ws
        self.job_uuid = job_uuid

    def publish(self, progress):
        data = {
            'action': 'publish',
            'topic':self.job_uuid,
            'content': progress
        }
        self.ws.send(json.dumps(data))

    def send(self, data:dict):
        self.publish(data)

    def error(self, data:dict):
        self.publish(data)
    

def send_progress(sender: Sender, data: dict):
    sender.send(data)

def send_error(sender: Sender, data: dict):
    sender.error(data) 

def compute_result(train_data_df:pd.DataFrame, test_data_df:pd.DataFrame, job_config:dict, thresholds:Thresholds, queue, config:dict):
    dependent_var_colname = get_node_unique_identifier(config['dependent_var']['pmModel'])
    threshold_instance = config['threshold_class'](dependent_variables=[dependent_var_colname], explanatory_vars=config['explanatory_var'], use_stats=config['stats'], model=config['model_class'])

    # # Set dependent var as value named col. since classes and outlier detection looking vor 'value'
    # train_data_df['value'] = train_data_df['value']
    # if test_data_df.empty == False:
    #     test_data_df['value'] = test_data_df['value']
    
    result = None
    try:          
        if job_config['trainMethod'] == 'isolatedWindows':                
            results_df = thresholds.train_static_data(threshold_instance, train_data_df, test_data_df)

        if job_config['trainMethod'] == 'movingWindows':
            results_df, _, _ = thresholds.train_rolling_data(threshold_instance, train_data_df, job_config['periodsDay']['train'], job_config['periodsDay']['retrain'], True)


        if not results_df.empty:
            mseEScaled = calculate_metrics(results_df, dependent_var_colname)
            # mseEScaled can be very high due to binning error in #45108
            # mseEScaled can be 0.0 due to most Y_pred = nan. Probably a consequent of to many missing infos (minBinElements)
            if mseEScaled > 10000 or mseEScaled != 0.0:
                missingDataCount = 'N/A'
                if hasattr(threshold_instance, 'missing_data'):
                    missingDataCount = len(threshold_instance.missing_data)
                if np.isnan(mseEScaled) == False:
                    result = {
                        'uuid': str(uuid.uuid4()),
                        'config': {
                            'dependentVar': config['dependent_var'],
                            'task': {
                                'name': 'maxpermissiblevalues',
                                'method': config['threshold_name'],
                                'model': config['model_name'],
                            },
                            'explanatoryVariables': config['explanatory_var_id'],
                            'missingDataCount': missingDataCount,
                            'metric': round(mseEScaled, 3),
                            'stats': config['stats'],
                            'tuningJobUuid': '',
                        }  
                    }
    except Exception as e:
        exc_type, _, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        logging.error(error)

    # put result to compute remaining percentages
    queue.put(result)
        
    return result          

def build_config_list(job, data_df):            
    job_config = job['config']
    dataframes = []
    for date in job_config['trainingDates']:
        dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
        dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
        dataframes.append(data_df[dt_start:dt_end])
    train_data_df = pd.concat(dataframes)

    test_data_df = pd.DataFrame()
    if len(job_config['measureDates']) > 0:
        dataframes = []
        for date in job_config['measureDates']:
            dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
            dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
            dataframes.append(data_df[dt_start:dt_end])
        test_data_df = pd.concat(dataframes)

    # Fetch and combine data
    thresholds = Thresholds()
    interval = job_config['interval']          
            
    m = multiprocessing.Manager()
    queue = m.Queue()                
                        
    independent_var_nodes = get_temporal_var_list(interval)
    if 'holiday' in data_df.columns:
        independent_var_nodes.append(get_feature_template('holiday'))
                        
    # Synchronizes mandatory flag between important features sets and nodes
    important_feature_sets = set_mandatory_flag(job_config['importantFeaturesSets'], job_config['nodes'])                
                                                        
    #----- Create combinations
    combinations = get_dependent_independent_vars_combinations(important_feature_sets)
    # ----- Create config list
    config_list = []
    for combination in combinations:
        for independent_vars in combination['independent_vars']:
            explanatory_var_name_list = get_colname_list(independent_vars)              
            
            for threshold_dict in METHODS_CLASS_DICT.items():
                if threshold_dict[0] == 'thresholdsGrouping':  
                    config_list.append((
                        train_data_df, 
                        test_data_df, 
                        job_config, 
                        thresholds, 
                        queue,
                        {
                            'dependent_var': combination['dependent_var'],
                            'threshold_name': threshold_dict[0],
                            'threshold_class': threshold_dict[1],
                            'model_name': '',
                            'model_class': '',                                
                            'explanatory_var': explanatory_var_name_list,
                            'explanatory_var_id': independent_vars,                        
                            'stats': 'mean', # for stats in ['median', 'mean']: // move to hyperparameter tuning   
                        }
                    ))     

                if threshold_dict[0] == 'thresholdsRegression':
                        for model in thresholds._regressor_model_list.items():
                            config_list.append((
                                train_data_df, 
                                test_data_df, 
                                job_config, 
                                thresholds, 
                                queue,
                                {
                                    'dependent_var': combination['dependent_var'],
                                    'threshold_name': threshold_dict[0],
                                    'threshold_class': threshold_dict[1],
                                    'model_name': model[0],
                                    'model_class': model[1](),
                                    'explanatory_var': explanatory_var_name_list,
                                    'explanatory_var_id': independent_vars,
                                    'stats': '',                                        
                                }
                            ))
        return config_list, queue



async def run_computation(response, cpus, config_list, queue, sender):
    pool = multiprocessing.Pool(processes=cpus)        

    #------- Calculate metrics         
    chunk_size = len(config_list) // cpus                   
    results = pool.starmap_async(compute_result, config_list, chunksize=chunk_size)

    # monitor task results
    while True:
        if results.ready():
            break
        else:
            computed_tasks = queue.qsize()
            response['computedConfigs'] = computed_tasks
            response['percentage'] = round((computed_tasks / len(config_list)) * 100)
            send_progress(sender, response)
            results.wait(1)                
    pool.close()
    pool.join()

    return results


def reduce_results(results, amount_results):
    #---- Split dependent variables with their metrics and reduce the amount of configs by max_configs_per_dependent_variable
    max_configs_per_dependent_variable = amount_results
    dependent_vars_results = {}
    for result in results.get():
        # If error occurs in task computation, result will equal to None.            
        if result != None:
            dependent_var_uuid = result['config']['dependentVar']['uuid']
            if dependent_var_uuid not in dependent_vars_results:
                dependent_vars_results[dependent_var_uuid] = deque([], maxlen=max_configs_per_dependent_variable)
            
            dependent_vars_result_configs = dependent_vars_results[dependent_var_uuid]
            if len(dependent_vars_result_configs) > 0:     
                dependent_vars_result_configs = deque(sorted(dependent_vars_result_configs, key=lambda x: x['config']['metric'], reverse=False), maxlen=max_configs_per_dependent_variable)
                if len(dependent_vars_result_configs) < max_configs_per_dependent_variable:
                    dependent_vars_result_configs.appendleft(result)
                elif dependent_vars_result_configs[-1]['config']['metric'] > result['config']['metric']:
                    dependent_vars_result_configs.appendleft(result)
            else:
                dependent_vars_result_configs.appendleft(result)
            dependent_vars_results[dependent_var_uuid] = dependent_vars_result_configs
            
    results = []                                                      
    for result_config in dependent_vars_results.items():
        results = results + list(result_config[1])
    
    #--- Sorting to have best metric at very top
    results.sort(key=lambda x: x['config']['metric'], reverse=False)

    return results


def get_input_data(spool_uuid_dir:str):
    input_file = open(f"{spool_uuid_dir}/input.json", encoding='utf-8')
    input_data = json.load(input_file)
    input_file.close()
    job = input_data['job']

    data_df = pd.DataFrame(input_data['dataFrame'])
    data_df = data_df.set_index('dateTime', drop=False)
    data_df.index = pd.to_datetime(data_df.index)

    return job, data_df   

# Task to handle manual ping/pong messages
async def ping_pong(ws):
    while True:
        try:
            await ws.ping()
        except Exception:            
            break
        await asyncio.sleep(10)


async def main(args, sender):
    start = time.time()
    amount_results = 50
    job_uuid = args.jobUuid
    response = {
        'jobUuid': job_uuid, 
        'status': 'computing',
        'percentage': 0, 
        'configListLength': 0, 
        'computedConfigs': 0,
        'totalDuration': '',
        'error': '',
    }

    if args.context == 'cluster':
        asyncio.create_task(ping_pong(ws))

    spool_uuid_dir = f"/opt/intelli-service/spool/{args.jobUuid}"        
    if args.context  == 'local':
        spool_uuid_dir =  f"spool/{args.jobUuid}"

    job, data_df = get_input_data(spool_uuid_dir)
    config_list, queue = build_config_list(job, data_df)
    response['configListLength'] = len(config_list)

    # send first status
    send_progress(sender, response)

    # Integrated asyncion to use poing ping in paralell
    computation_task = asyncio.create_task(run_computation(response, job['cpu'], config_list, queue, sender))

    # Wait for the send task to complete
    results = await computation_task

    reduced_results = reduce_results(results, amount_results)
    with open(f"{spool_uuid_dir}/output.json", 'w', encoding='utf-8') as outfile:
        json.dump(reduced_results, outfile)            

    end = time.time()        
    response['computedConfigs'] = len(config_list)
    response['totalDuration'] = str(timedelta(seconds=math.ceil(end - start))) 
    response['status'] = 'done'
    response['percentage'] = 100

    send_progress(sender, response)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--jobUuid', help='Uuid of the page to load')
    parser.add_argument('--context', help='execution target. ie. local, cluster')
    parser.add_argument('--token', help='Security token for services')
    parser.add_argument('--processManager', help='Base url of the process manager')
    args = parser.parse_args()

    sender = None
    try:
        if args.context == 'local':
            sender = LocalSender()
        else:
            ws = connect("ws://intelli-controller.intelli.svc.cluster.local/v1/broker")
            sender = WebSocketSender(ws, args.jobUuid)

        asyncio.run(main(args, sender))
    except BaiException as be:
        error = "{}".format(be)
        response = {
            'jobUuid': args.jobUuid,
            'status': 'stopped',
            'error': error
        }
        send_error(sender, response)
    except Exception as e:
        exc_type, _, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        response = {
            'jobUuid': args.jobUuid,
            'status': 'stopped',
            'error': error
        }        
        send_error(sender, response)    