import pandas as pd
from sklearn.preprocessing import StandardScaler 
import numpy as np

def get_alarms(data:pd.DataFrame, lookback_horizon:int, no_events_threshold:int):
    """Raise alarms based on too many events 
    
    :param data pd.DataFrame containing events, i.e. with column 'event'
    :param lookback_horizon  window in the past to check for events (CAVEAT: given in number of samples *not* Delta_time)
    :param no_events_threshold  number of events in the lookback_horizon that should trigger an alarm
    :return: 
    """
    adata = data[['event']].copy()
    adata = adata.rolling(lookback_horizon).sum()
    return (adata['event'] >= no_events_threshold) & (data['event'] == True)

def calculate_metrics(results_df:pd.DataFrame, 
                      lookback_horizon:int, 
                      no_events_threshold:int,
                      punish_missing_values=100) -> pd.DataFrame:
    """Calculate events, alarms, metrics 
    :param pd.DataFrame results_df:  DataFrame containing the results from the function do_backtesting
    :param int lookback_horizon:  Number of past days to check for an event in order to trigger an alarm
    :param int no_events_threshold:  Number of events in the lookback_horizon that will trigger an alarm
    :return tuple: (pd.DataFrame, dict) Where:
        1. Return results_df with xxxx added columns originating from the 'thresholding class' during backtesting
            * 'event'      : True if the min_dist > 0.0, i.e. if the value is outside of the thresholds
            * 'alarm'      : True, see function `get_alarms` 
            * 'error'      : Distance between predicted (`Y_pred`) and measured value (`value`) 
            * 'errorE'     : "Extended Error" -> error but all positions without an error are filled with max(error)
            * 'SqrError'   : squared error
            * 'SqrErrorE'  : squared errorE
            * 'rollingMSE' : moving mean over sqrError with window length `lookback_horizon`
            * 'rollingMSEE': moving mean over sqrErrorE with window length `lookback_horizon`
            * 'rollingEventCount' : number events summed up over `lookback_horizon` 
        2. dict of summary statistics
    """
    idx = results_df['min_dist'] > 0.0
    results_df['event'] = idx
    
    # calculate alarms
    results_df['alarm'] = get_alarms(results_df, lookback_horizon=lookback_horizon, no_events_threshold = no_events_threshold)
    
    # (squared) error 
    results_df['error'] = results_df['value'] - results_df['Y_pred']
    results_df['SqrError'] = np.square(results_df['error'])
    
    # "error Extended" -> squared error also for positions withhout Threshold
    max_abs_error = results_df['error'].abs().max()
    idx_nothreshold = results_df['upper'] == np.inf
    results_df['errorE'] = results_df['error'].copy()
    
    # Weighting the missing thresholds with punish_missing_values times the max. error of the the
    results_df.loc[idx_nothreshold, 'errorE'] =  punish_missing_values * max_abs_error 
    results_df['SqrErrorE'] = np.square(results_df['errorE'])
    
    # "Normalized error Extended"  
    stdScaler = StandardScaler().fit(results_df['value'].values.reshape(-1, 1))
    scaled_value = stdScaler.transform(results_df['value'].values.reshape(-1, 1))
    scaled_prediction = stdScaler.transform(results_df['Y_pred'].values.reshape(-1, 1))   
    
    results_df['valueScaled'] = scaled_value
    results_df['Y_predScaled'] = scaled_prediction
    results_df['errorScaled'] = scaled_value - scaled_prediction 
    results_df['SqrErrorScaled'] = np.square(results_df['errorScaled'])
    
    # "error Extended" -> squared error also for positions without Threshold
    mean_abs_error_scaled = results_df['errorScaled'].abs().max()
    results_df['errorEScaled'] = results_df['errorScaled'].copy()
    results_df.loc[idx_nothreshold, 'errorEScaled'] = punish_missing_values * mean_abs_error_scaled
    results_df['SqrErrorEScaled'] = np.square(results_df['errorEScaled'])  
    
    # overall Statistics 
    # likelyhood of error
    # gm = GaussianMixture(n_components=1, verbose=True).fit( results_df['error'].values.reshape(-1, 1) )
    # # print(gm.means_, gm.covariances_)
    # results_df['gaussLogLikelyhood'] = gm.score_samples(results_df['error'].values.reshape(-1,1))
    # windowed calculations
    results_df[f"rollingMSE"] = results_df['SqrError'].rolling(window=pd.to_timedelta(lookback_horizon, unit='days')).mean()
    results_df[f"rollingMSEE"] = results_df['SqrErrorE'].rolling(window=pd.to_timedelta(lookback_horizon, unit='days')).mean()
    results_df[f"rollingMSEScaled"] = results_df['SqrErrorScaled'].rolling(window=pd.to_timedelta(lookback_horizon, unit='days')).mean()
    results_df[f"rollingMSEEScaled"] = results_df['SqrErrorEScaled'].rolling(window=pd.to_timedelta(lookback_horizon, unit='days')).mean()
    results_df[f"rollingEventCount"] = results_df['event'].rolling(window=pd.to_timedelta(lookback_horizon, unit='days')).sum()
    # results_df[f"rollingSumLikelyhood"] = results_df['gaussLogLikelyhood'].rolling(window=pd.to_timedelta(LOOKBACK_HORIZON, unit='days')).sum()
    
    # summary_metrics
    summary_metrics = {}
    summary_metrics['mse'] = results_df['SqrError'].mean() 
    summary_metrics['mseE'] = results_df['SqrErrorE'].mean()
    summary_metrics['mseScaled'] = results_df['SqrErrorScaled'].mean()
    summary_metrics['mseEScaled'] = results_df['SqrErrorEScaled'].mean()
    summary_metrics['maxRollingMSE'] =  results_df['rollingMSE'].max()
    summary_metrics['maxRollingMSEE'] =  results_df['rollingMSEE'].max()
    summary_metrics['maxRollingMSEScaled'] =  results_df['rollingMSEScaled'].max()
    summary_metrics['maxRollingMSEEScaled'] =  results_df['rollingMSEEScaled'].max()
    error_stats_df = results_df['error'].describe()
    summary_metrics['meanError'] = error_stats_df['mean']
    summary_metrics['medianError'] =  error_stats_df['50%']
    summary_metrics['stdError'] =  error_stats_df['std']
    error_stats_df = results_df['errorScaled'].describe()
    summary_metrics['meanErrorScaled'] = error_stats_df['mean']
    summary_metrics['medianErrorScaled'] =  error_stats_df['50%']
    summary_metrics['stdScaled'] =  error_stats_df['std']
    r = (results_df.index[-1] - results_df.index[0]).days
    summary_metrics['relNanThresholds'] = np.sum(results_df['upper'] == np.inf) / r
    summary_metrics['relEvents'] = np.sum(results_df['min_dist'] > 0.0) / r
    summary_metrics['relAlarms'] = np.sum(results_df['alarm']) / r
    
    return results_df, summary_metrics
