import pandas as pd
import numpy as np
from copy import copy
from scipy import stats
from datetime import datetime, timedelta

def preprocessing(df, time_col='Zeit', target_col='Wert'):
    # copy incomming dataframe
    df = copy(df)

    # Set format for performance boost: https://stackoverflow.com/a/32034914
    # df = df.set_index(pd.to_datetime(df[time_col], format="%Y-%m-%d %H:%M"))
    # df.drop(columns='x', inplace=True) # For some reason x not being dropped with set_index

    # drop unused columns
    unused_cols = list(set(df.columns) - set([target_col]))
    df = df.drop(columns=unused_cols)

    # calculate helper columns
    df['year'] = [str(x).zfill(2) for x in  df.index.year]
    df['month'] = [str(x).zfill(2) for x in  df.index.month]
    df['week'] = [str(x).zfill(2) for x in  df.index.isocalendar().week]
    df['weekday'] = [str(x) for x in  df.index.weekday]
    df['day'] = [str(x).zfill(2) for x in  df.index.day]
    df['week_of_year'] = df['year'] + df['week']
    df['day_of_year'] = df['year'] + df['month'] + df['day']
    return df

def timedelta_from_str(start, end):
    s = pd.to_datetime(start)
    e = pd.to_datetime(end)
    d = (e-s).seconds/3600
    return d

def exponential_moving_average(x_1, x_0, alpha):
    return (alpha*x_1 + (1-alpha)*x_0)

def get_night_time(X, verbose):
    # copy incomming dataframe
    X = copy(X)

    # hour freq, set to 4 for 15 mins freq
    freq_factor = 1

    # small median filter to improve edge detection
    med_2h = X.rolling(freq_factor*2, center=True).median()
    # big rolling mean filter to track the boundary of ON and OFF events
    mean_48h = X.rolling(freq_factor*48, center=True).mean()
    # extract ON and OFF events and convert them to 1/0
    edges = (med_2h > mean_48h).astype(int)
    # detect edges. 1 meaning switch-ON, -1 meaning switch-OFF
    edges = edges.diff().fillna(0)
    # extract all switch-ON and switch-OFF events
    end = edges[edges > 0]
    start = edges[edges < 0]

    if len(start) == 0 and len(end) == 0:
        raise Exception("monitoringTimeGuessedFailed")

    # convert associated timestamps to minutes (from 00:00h, local time)
    start_events_minutes = start.index.map(lambda x: x.hour*60 + x.minute)
    end_events_minutes = end.index.map(lambda x: x.hour*60 + x.minute)
    # find the peak (statistical mode) of the distribution
    start_minutes = stats.mode(start_events_minutes)[0][0]
    end_minutes = stats.mode(end_events_minutes)[0][0]

    # check validity of estimate
    if start_minutes == end_minutes:
        return 0,0,0

    # shrink boundaries by +- 1h
    if (start_minutes + 60) >= 24*60:
        start_minutes = 24*60 - (start_minutes + 60)
    else:
        start_minutes = start_minutes + 60
    if (end_minutes - 60) <= 0:
        end_minutes = 24*60 + (end_minutes - 60)
    else:
        end_minutes = end_minutes - 60

    # calculate delta
    if start_minutes >= end_minutes:
        delta_hours = 24 - (start_minutes - end_minutes)/60
    if end_minutes > start_minutes:
        delta_hours = (end_minutes - start_minutes)/60

    if verbose:
        print('Estimated day/night border: ')
        print('\t Night-Start: {}'.format(minutes_to_timestamp(start_minutes)))
        print('\t Night-End: {}'.format(minutes_to_timestamp(end_minutes)))
        print('\t Delta: {}h'.format(delta_hours))

    return start_minutes, end_minutes, delta_hours

def minutes_to_timestamp(minutes):
    return '{:02d}:{:02d}'.format(0 if (minutes//60)==24 else minutes//60, minutes%60)

def check_time_estimate(night_start_minutes, night_end_minutes, night_delta_hours, default_night_start, default_night_end, verbose):
    # sanity check the delta
    if (night_delta_hours > 20) | (night_delta_hours < 2):
        print('Error: Time estimate unreliable!')
        print('\t Using default timeframe')
        night_start = default_night_start
        night_end = default_night_end
        night_delta_hours = timedelta_from_str(start=default_night_start, end=default_night_end)
    else:
        night_start = minutes_to_timestamp(night_start_minutes)
        night_end = minutes_to_timestamp(night_end_minutes)
        night_delta_hours = night_delta_hours

    if verbose:
        print('Final day/night border used:')
        print('\t Start: {}'.format(night_start))
        print('\t End: {}'.format(night_end))
        print('\t Delta: {}h'.format(night_delta_hours))


    return night_start, night_end, night_delta_hours

def calculate_threshold(monitoring_time, th_percent, min_th):
    monitoring_time_median = np.median(monitoring_time)
    th = monitoring_time_median + monitoring_time_median * th_percent

    th = np.max([th, min_th])
    return th


def get_event_annotations(events, daily_event_count, week_event_count, start_time, end_time):
  df = pd.DataFrame(events)
  event_dates = df['x'].values

  daily_alarms = []
  weekly_alarms = []

  for date in event_dates:
    day_of_year = datetime.strptime(date, "%Y-%m-%d")

    # Get daily alarms
    daily_list = [(day_of_year - timedelta(days=x)).strftime("%Y-%m-%d") for x in range(daily_event_count)]
    if all(item in event_dates for item in daily_list):
      daily_alarms.append(daily_list[0])

    # Get weekly alarms excluding sundays
    weekly_list = [(day_of_year - timedelta(weeks=x)).strftime("%Y-%m-%d") for x in range(week_event_count)]
    if all(item in event_dates for item in weekly_list):
      weekly_alarms.append(weekly_list[0])

  alarms_dw = set(weekly_alarms).intersection(daily_alarms)
  daily_alarms = set(daily_alarms).difference(alarms_dw)
  weekly_alarms = set(weekly_alarms).difference(alarms_dw)


  # Build alarms accorind to apexchart annotation api
  annotationTemplate = {
      'borderColor': '#FB441D',
      'fillColor': '#FB441D',
      'label': {
        'borderColor': '#FB441D',
        'style': {
          'background': '#FB441D',
          'color': '#fff'
        }
      }
    }

  annotations = []
  for date in list(daily_alarms):
    annotation = annotationTemplate.copy()
    xDate, x2Date = getAnnotationDates(date, start_time, end_time)
    annotation['x'] = xDate
    annotation['x2'] = x2Date
    annotation['label']['text'] = 'D'
    annotations.append(annotation)

  weekly_alarms = []
  for date in list(weekly_alarms):
    annotation = annotationTemplate.copy()
    xDate, x2Date = getAnnotationDates(date, start_time, end_time)
    annotation['x'] = xDate
    annotation['x2'] = x2Date
    annotation['label']['text'] = 'W'
    annotations.append(annotation)

  alarms_dw = []
  for date in list(alarms_dw):
    annotation = annotationTemplate.copy()
    xDate, x2Date = getAnnotationDates(date, start_time, end_time)
    annotation['x'] = xDate
    annotation['x2'] = x2Date
    annotation['label']['text'] = 'DW'
    annotations.append(annotation)

  return {'xaxis': annotations}

def getAnnotationDates(date, start_time, end_time):
  date_format = '%Y-%m-%d %H:%M'
  datetime_start = datetime.strptime("{} {}".format(date, start_time), date_format)
  datetime_end = datetime.strptime("{} {}".format(date, end_time), date_format)

  xDate = datetime_start.timestamp()*1000
  x2Date = datetime_end.timestamp()*1000

  if start_time > end_time:
      x2Date =(datetime_end + timedelta(days=1)).timestamp()*1000
  return xDate, x2Date


def get_config_mappig(index):
  configMapping = {
    1: {
      'lookback': 2,
      'thresholdPercent': 0.81,
      'alpha': 0.81,
    },
    2: {
      'lookback': 3,
      'thresholdPercent': 0.72,
      'alpha': 0.72,
    },
    3: {
      'lookback': 5,
      'thresholdPercent': 0.72,
      'alpha': 0.72,
    },
    4: {
      'lookback': 6,
      'thresholdPercent': 0.63,
      'alpha': 0.63,
    },
    5: {
      'lookback': 7,
      'thresholdPercent': 0.54,
      'alpha': 0.54,
    },
    6: {
      'lookback': 8,
      'thresholdPercent': 0.45,
      'alpha': 0.45,
    },
    7: {
      'lookback': 9,
      'thresholdPercent': 0.36,
      'alpha': 0.36,
    },
    8: {
      'lookback': 22,
      'thresholdPercent': 0.27,
      'alpha': 0.27,
    },
    9: {
      'lookback': 12,
      'thresholdPercent': 0.18,
      'alpha': 0.18,
    },
    10: {
      'lookback': 13,
      'thresholdPercent': 0.09,
      'alpha': 0.09,
    }
  }
  return configMapping[index]