export default function useComponent () {
  let parent = null
  
  function setParent(p) {
    parent = p 
  }

  function append(name, slot = () => []) {
    this.children = this.children ? this.children : {}
    const slotName = typeof name === 'function' ? 'default' : name
    const slotFunc = typeof name === 'function' ? name : slot
    this.children[slotName] = slotFunc
    return this
  }

  function on(event, callback = () => {}) {
    this.props = this.props ? this.props : {}
    if (typeof event === 'object' && event !== null) {
      const eventNames = Object.keys(event)
      this.props.clickable = eventNames.find((name) => name.includes('Click')) ? true : this.props.clickable  
      this.props = { ...this.props, ...event }
    } else {
      this.props[event] = callback
      this.props.clickable = event.includes('Click') ? true : this.props.clickable  
    }
    return this
  }

  function getToken() {
    return parent.ctx.httpClient.getTokenString()
  }

  function socket(url, callbacks = { connect(){}, message(){}, error(){}, close(){}}) {
    url = url.startsWith('ws') ? url : parent.ctx.endpoints.getWs(url)
    let s = new WebSocket(url)
  
    s.onopen = function(e) {
      callbacks.connect(e)
    }
    
    s.onmessage = function(event) {
      try {
        callbacks.message(JSON.parse(event.data))
      } catch (e) {
        callbacks.message(event.data)
      }
    }
    
    s.onclose = function(event) {
      if (event.wasClean) {
        callbacks.close(event)
      } else {
        console.error('[close] Connection died', event);
      }
    }
    
    s.onerror = function(err) {
      callbacks.error(err)
    }

    function send(data) {
      try {
        if (typeof data === 'object') {
          s.send(JSON.stringify(data))
        } else {
          s.send(data)
        }
      } catch (e) {
        console.error(e)
      }
    }

    function close(code, reason) {
      try {
        s.close(code, reason)
      } catch (e) {
        console.error(e)
      }
    }
    
    return {
      close,
      send,
    }
  }

  function createTableColsByObject(obj, unsortable) {
    const keys = Object.keys(obj)
    return keys.map((key) => ({ name: key, i18nLabel: key, field: key, sortable: !unsortable, key }))
  }

  function attrs(props = {}) {
    const fn = on.bind(this)
    return fn(props)
  }

  function createDefaultPagination(rowsPerPage = 0) {
    return {
      sortBy: 'desc',
      descending: false,
      page: 1,
      rowsPerPage,
    }
  }

  function renderIf(state, component) {
    return state ? component : ''
  }

  function router() {
    return {
      push: (path, query, meta) => parent.ctx.routing.push(query ? { path, query, meta } : path),
      replace: (query) => parent.ctx.routing.replace({ query: { ...parent.ctx.routing.route.query, ...query }}),
      back: () => parent.ctx.routing.go(-1),
      refresh: () => parent.ctx.routing.go(0),
      query: parent.ctx.routing.route.query,
      params: parent.ctx.routing.route.params,
      meta: parent.ctx.routing.route.meta,
    }
  }

  function notify() {
    return {
      saved: () => {
        parent.ctx.notify.plugin.positive(parent.ctx.i18n.t('saved'))
      },
      added: (name) => {
        parent.ctx.notify.plugin.positive(name ? parent.ctx.i18n.t('namedAdded', { name }) : parent.ctx.i18n.t('added'))
      },
      positive: (text) => {
        let msg = text
        if (typeof text === 'object' && !Array.isArray(text)) {
          if (text !== null && text.data && text.data.message) {
            msg = text.data.message
          }
        }
        parent.ctx.notify.plugin.positive(msg)
      },
      negative: (text) => {
        let msg = text
        if (text !== null && typeof text === 'object' && !Array.isArray(text)) {
          if (text.response && text.response.data && text.response.error) {
            msg = text.response.error
          } else if (text.response && text.response.data && text.response.message) {
            msg = text.response.message
          } else {
            msg = text.message
          }   
        }
        parent.ctx.notify.plugin.error(msg)
      },
    }
  }

  function validRoles(permissions = []) {
    if (permissions.includes('*')) {
      return true
    }
   return parent.ctx.user.hasPermissions(permissions) 
  }

  function getDateString(ts) {
    return ts ? new Date(ts).toLocaleString() : ''
  }
  
  // Collection helper for push, remove, moveUp or moveDown, sort a element into a array.
  function collection() {
    return {
      add: (items, item) => items.push(item),
      remove: (items, item) => {
        const idx = items.indexOf(item)
        if (idx > -1) {
          items.splice(idx, 1)
        }
      },
      sortByKey: (items, key) => {
        items.sort((a, b) => {
              const fa = a[key].toLowerCase();
              const fb = b[key].toLowerCase();
              if (fa < fb) {
                  return -1;
              }
              if (fa > fb) {
                  return 1;
              }
              return 0;
          })
      },
      moveUp: (items, item) => {
        const idx = items.indexOf(item)
        if (idx > -1) {
          items.splice(idx, 1)
        } 
        const cols = items.splice(idx, 1)
        items.splice(idx - 1, 0, cols[0])
      },
      moveDown: (items, item) => {
        const idx = items.indexOf(item)
        if (idx > items.length) {
          return
        }
        const cols = items.splice(idx, 1)
        items.splice(idx + 1, 0, cols[0])
      }
    }
  }

  function clickStop() {
    return () => parent.ctx.withModifiers(() => {}, ['stop'])
  }
  
  const context = {
    clickStop,
    utils: {
      collection,
    },
    validRoles,
    router,
    notify,
    renderIf,
    socket,
    getToken,
    setParent,
    createTableColsByObject,
    createDefaultPagination,
    getDateString,
    getObjCopy: (data) => JSON.parse(JSON.stringify(data)),
    classSpacing: (spaceNames) => Array.isArray(spaceNames) ? spaceNames.map((name) => `q-${name}`) : `q-${spaceNames}`,
    classCell: (cells) => cells.map((name) => `q-${name}`),
    class: (c) => c.map((name) => `q-${name}`),
    classTableGrid: () => 'q-pa-xs col-xs-12 col-sm-6 col-md-4 col-lg-3 grid-style-transition',
    element: (element) => ({
      type: element,
      props: {
      },
      append,
      attrs,
      on,
    }),
    div: () => ({
      type: 'section',
      props: {
      },
      append,
      attrs,
      on, 
    }),
    avatar: (text, color = 'dark', textColor = 'white', size = '26px') => ({
      type: 'NgAvatar',
      props: {
        color,
        textColor,
        size,
      },
      children: {
        default: () => [
          typeof text === 'string' && text.length > 2 ? text.substr(0, 2).toUpperCase() : text,
        ]
      },
      append,
      attrs,
      on,    
    }),
    breadcrumbs: (items = []) => ({
      type: 'NgBreadcrumbs',
      props: {
        items: [
          {
            icon: 'apps',
            to: '/',
          },
          ...items,
        ],
        class: 'q-mt-sm q-mb-xs q-mx-sm',
        gutter: 'sm'
      },
      append,
      attrs,
      on,
    }),
    page: (permissions = [], onLoggedIn = () => {}, onLoggedOut = () => {}, authDisabled = true) => ({
      type: 'NgPage',
      props: {
        dense: true,
        permissions: typeof permissions === 'function' ? [] : permissions,
        authDisabled,
        onLoggedIn: typeof permissions === 'function' ? permissions : onLoggedIn,
        onLoggedOut: typeof permissions === 'function' && onLoggedIn ? onLoggedIn : onLoggedOut,
        class: 'q-ma-none q-pa-none'
      },
      append,
      attrs,
      on,
    }),
    spacing: (spaceNames = ['ma-sm']) => ({
      type: 'div',
      props: {
        class: spaceNames.map((name) => `q-${name}`),
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    separator: (inset) => ({
      type: 'div',
      props: {
        inset,
      },
      children: {
        default: () => [
          {
            type: 'NgSeparator',
            props: {
              inset,
            },
          },
        ]
      },
      append,
      attrs,
      on,
    }),
    text: (type, text) => ({
      type: 'NgText',
      props: {
        type: text === undefined ? 'p' : type,
        text: text === undefined ? type : text,
      },
      append,
      attrs,
      on,
    }),
    button: (label, icon, color) => ({
      type: 'NgActionButton',
      props: {
        label,
        color,
        icon: typeof icon === 'function' ? undefined : icon,
        onClick: typeof icon === 'function' ? icon : () => ({}),
      },
      append,
      attrs,
      on,
    }),
    buttonFlat: (label, icon) => ({
      type: 'NgActionButton',
      props: {
        label,
        icon,
        flat: true,
      },
      append,
      attrs,
      on,
    }),
    buttonFlatLeft: (label, icon) => ({
      type: 'NgActionButton',
      props: {
        label,
        icon,
        flat: true,
        align: 'left',
        padding: 'none',
      },
      append,
      attrs,
      on,
    }),
    buttonGroup: (label) => ({
      type: 'NgActionGroup',
      props: {
        label,
        rounded: true,
      },
      append,
      attrs,
      on,
    }),
    chip: (label, color) => ({
      type: 'NgChip',
      props: {
        label,
        color,
      },
      append,
      attrs,
      on,
    }),
    dialog: (title, value, onUpdate = () => {}, size = 'medium') => ({
      type: 'NgDialog',
      props: {
        title,
        draggable: true,
        maximizing: true,
        maximized: !parent.ctx.platform.browser,
        modelValue: value, // assign value from data object with this.[variableName]
        'onUpdate:modelValue': onUpdate, // assign value from data object with update value
        noActions: true,
        small: size === 'small',
        medium: size === 'medium',
        large: size === 'large',
      },
      children: {
        default: () => [],
      },
      append,
      attrs,
      on,
    }),
    dialogConfirm: (text, value, onUpdate = () => {}) => ({
      type: 'NgDialogConfirm',
      props: {
        text,
        draggable: true,
        maximizing: false,
        modelValue: value, // assign value from data object with this.[variableName]
        'onUpdate:modelValue': onUpdate, // assign value from data object with update value
        noAction: true,
      },
      children: {
        default: () => [],
      },
      append,
      attrs,
      on,
    }),
    dialogDelete: (text, value, onUpdate = () => {}) => ({
      type: 'NgDialogDelete',
      props: {
        text,
        draggable: true,
        maximizing: false,
        modelValue: value, // assign value from data object with this.[variableName]
        'onUpdate:modelValue': onUpdate, // assign value from data object with update value
        noAction: true,
      },
      children: {
        default: () => [],
      },
      append,
      attrs,
      on,
    }),
    input: (label, type, value, onUpdate = () => {}) => ({
      type: 'NgFormInput',
      props: {
        type,
        label,
        name,
        modelValue: value, // assign value from data object with this.[variableName]
        'onUpdate:modelValue': onUpdate, // assign value from data object with update value
        dense: true,
        hint: undefined
      },
      append,
      attrs,
      on,
    }),
    groups: () => ({
      type: 'NgFormGroups',
      props: {
      },
      append,
      attrs,
      on,    
    }),
    group: (title) => ({
      type: 'NgFormGroupItem',
      props: {
        title,
      },
      append,
      attrs,
      on,    
    }),
    form: (title) => ({
      type: 'NgForm',
      props: {
        title,
        column: true,
        submitLabel: 'save',
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    formWithBack: (title) => ({
      type: 'NgForm',
      props: {
        title,
        column: true,
        submitLabel: 'save',
        showReset: true,
        resetLabel: 'back',
        onReset: () => parent.ctx.routing.go(-1),
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    formDialog: (title, value, onUpdate = () => {}, formUiKey = '', size = 'medium') => {
      const Form = window.nguiAsyncImport(formUiKey)
      const form = {
        type: Form,
        props: {
        },
        append,
        attrs,
        on,
      }
      return {
        type: 'NgDialog',
        props: {
          title,
          draggable: true,
          maximizing: true,
          maximized: !parent.ctx.platform.browser,
          modelValue: value, // assign value from data object with this.[variableName]
          'onUpdate:modelValue': onUpdate, // assign value from data object with update value
          noActions: true,
          small: size === 'small',
          medium: size === 'medium',
          large: size === 'large',
        },
        children: {
          default: () => [
            form
          ],
        },
        append,
        attrs,
        on,
        formAttrs: function (attrs) {
          form.attrs(attrs)
          return this
        },
        formOn: function (on) {
          form.on(on)
          return this    
        }
      }
    },
    badge: (label, icon, color = 'grey') => ({
      type: 'NgBadge',
      props: {
        icon,
        color,
      },
      children: {
        default: () => [
          parent ? parent.ctx.i18n.t(label) : label
        ]
      },
      append,
      attrs,
      on,
    }),
    listVirtual: (title, subtitle, items) => ({
      type: 'NgListVirtual',
      props: {
        title,
        subtitle: !items ? '' : subtitle,
        items: !items ? subtitle : items,
        separator: true,
        bordered: true,
        rounded: true,
        class: 'ellipsis',
        wrap: true,
      },
      append,
      attrs,
      on,
    }),
    listVirtualScroll: (title, subtitle, items, maxHeight = '70vh') => ({
      type: 'NgListVirtualScroll',
      props: {
        title,
        subtitle: !items ? '' : subtitle,
        items: !items ? subtitle : items,
        separator: true,
        bordered: true,
        rounded: true,
        maxHeight,
        wrap: true,
      },
      append,
      attrs,
      on,
    }),
    list: (title, subtitle) => ({
      type: 'NgList',
      props: {
        title,
        subtitle,
        dense: true,
        separator: true,
        bordered: false,
        rounded: true,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    item: (title, subtitle, expandable) => ({
      type: expandable ? 'NgListExpandItem' : 'NgListItem',
      props: {
        title,
        subtitle,
        dense: true,
        clickable: false,
      },
      children: {
        default: () => [
        ]
      },
      append,
      attrs,
      on,
    }),
    itemSection: () => ({
      type: 'NgListSection',
      props: {
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    label: (label) => ({
      type: 'NgListLabel',
      props: {
        label,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    caption: (label) => ({
      type: 'NgListLabel',
      props: {
        label,
        caption: false,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    header: (label) => ({
      type: 'NgListLabel',
      props: {
        label,
        header: true,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    expandItem: (title, subtitle, group) => ({
      type: 'NgListExpandItem',
      props: {
        title,
        subtitle,
        group,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    icon: (name = 'arrow_forward', color, size) => ({
      type: 'NgIcon',
      props: {
        name,
        color,
        size,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,    
    }),
    flex: (justify = 'start', items = 'center', gutter = 'md', wrap = 'no-wrap') => ({
      type: 'div',
      props: {
        class: `flex justify-${justify} items-${items} q-gutter-${gutter} ${wrap}`,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    grid: (justify = 'start', items = '', gutter = '') => ({
      type: 'NgGrid',
      props: {
        justify,
        gutterX: 'md',
        gutterY: 'md',
        items,
        gutter,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    column: (justify = 'start', items = '', gutter = '') => ({
      type: 'NgGrid',
      props: {
        justify,
        gutterX: gutter || 'md',
        gutterY: gutter || 'md',
        column: true,
        items,
        gutter,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    cell: (cellSpan = {col: 12, md: 6}) => {
      if (typeof cellSpan === 'number') {
        cellSpan = {
          col: 12,
          md: cellSpan
        }
      }
      
      return {
        type: 'NgGridItem',
        props: {
          ...cellSpan,
        },
        children: {
          default: () => []
        },
        append,
        attrs,
        on,
      }
    },
    table: (title, columns, rows = []) => {
      columns = columns.map((col) => col.align ? col : { ...col, ...{ align: 'left'}})
      return {
        type: 'NgTable',
        props: {
          columns,
          rows,
          title,
          grid: !parent.ctx.platform.browser,
          flat: true,
        },
        children: {
          default: () => [],
        },
        append,
        attrs,
        on,
        pagination(p, onUpdate = () => {}) {
          this.props.pagination = p
          this.props['onUpdate:pagination'] = onUpdate
          return this
        }
      }
    },
    tableGrid: (title, columns, rows = []) => {
      columns = columns.map((col) => col.align ? col : { ...col, ...{ align: 'left'}})
      return {
        type: 'NgTable',
        props: {
          columns,
          rows,
          title,
          flat: false,
          grid: true
        },
        children: {
          default: () => [],
        },
        append,
        attrs,
        on,
        pagination(p, onUpdate = () => {}) {
          this.props.pagination = p
          this.props['onUpdate:pagination'] = onUpdate
          return this
        }
      }
    },
    td: (props, modifiers, key) => ({
      type: 'NgTableTd',
      props: {
        props,
        key,
        onClick: modifiers ? parent.withModifiers(() => {}, Array.isArray(modifiers) ? modifiers : [modifiers]) : void 0,
      },
      append,
      attrs,
      on,
    }),
    tr: (props, modifiers) => ({
      type: 'NgTableTr',
      props: {
        props,
        onClick: modifiers ? parent.withModifiers(() => {}, Array.isArray(modifiers) ? modifiers : [modifiers]) : void 0,
      },
      append,
      attrs,
      on,
    }),
    th: (props, modifiers, key) => ({
      type: 'NgTableTh',
      props: {
        props,
        key,
        onClick: modifiers ? parent.withModifiers(() => {}, Array.isArray(modifiers) ? modifiers : [modifiers]) : void 0,
      },
      append,
      attrs,
      on,
    }),
    tdItem: (card) => {
      const container = {
        type: 'div',
        props: {
          class: 'q-pa-xs col-xs-12 col-sm-6 col-md-4 col-lg-3 grid-style-transition',
        },
        children: {
            default: () => [
              card,
            ],
          },
        append,
        attrs,
        on,
      }
      return container
    },
    menu: (label, icon, items) => ({
      type: 'NgMenu',
      props: {
        label,
        icon: icon === undefined ? 'more_vert' : icon,
        flat: Array.isArray(label) || label === undefined,
        dense: true,
        items: Array.isArray(label) ? label : items,
        class: 'edit-menu'
      },
      append,
      attrs,
      on,
    }),
    toolbarMenu: (label, items) => ({
      type: 'NgMenu',
      props: {
        label,
        flat: true,
        color: 'white',
        dense: false,
        items: items,
        noCaps: true,
      },
      append,
      attrs,
      on,
    }),
    defaultMenu: (label, icon, permissions = { update: [], delete: []}) => ({
      type: 'NgMenu',
      props: {
        label,
        icon: icon === undefined ? 'more_vert' : icon,
        flat: label === undefined,
        dense: true,
        round: label === undefined,
        enableGlobalClick: true,
        class: 'edit-menu',
        items: [
          {
            id: 'edit',
            label: 'edit',
            permissions: permissions.update,
          },
          {
            id: 'delete',
            label: 'delete',
            permissions: permissions.delete,
          }
        ]
      },
      append,
      attrs,
      on,
    }),
	workspaceMenu: (label, icon, permissions = { update: [], delete: [], read: []}) => ({
      type: 'NgMenu',
      props: {
        label,
        icon: icon === undefined ? 'more_vert' : icon,
        flat: label === undefined,
        dense: true,
        round: label === undefined,
        enableGlobalClick: true,
        class: 'edit-menu',
        items: [
          {
            id: 'edit',
            label: 'edit',
            permissions: permissions.update,
          },
          {
            id: 'delete',
            label: 'delete',
            permissions: permissions.delete,
          },
		  {
            id: 'export',
            label: 'export',
            permissions: permissions.read,
          }
        ]
      },
      append,
      attrs,
      on,
    }),
    customMenu: (label, icon, items = ['play', 'stop', 'restart']) => ({
      type: 'NgMenu',
      props: {
        label: Array.isArray(label) ? undefined : label,
        icon: Array.isArray(label) || icon === undefined ? 'more_vert' : icon,
        flat: Array.isArray(label) || label === undefined,
        dense: true,
        round: Array.isArray(label) || label === undefined,
        class: 'edit-menu',
        enableGlobalClick: true,
        items: (Array.isArray(label) ? label : items).map((item) => ({
          id: item,
          label: item
        })),
      },
      append,
      attrs,
      on,
    }),
    bar: (title) => ({
      type: 'NgMenubar',
      props: {
        title: title,
        menus: [],
        rounded: false,
        color: 'primary',
      },
      append,
      attrs,
      on,
      createMenu(title, textColor, items = [], permissions = [], hidden = false) {
        if (hidden || (typeof permissions === 'boolean' && permissions)) {
          return this
        }
        this.props.menus.push({ 
          id: title, 
          label: title, 
          textColor: Array.isArray(textColor) ? 'white' : textColor, 
          flat: true, 
          items: Array.isArray(textColor) ? textColor : items,
          permissions: Array.isArray(textColor) ? items : permissions,
          hidden: true,
        })
        return this
      }
    }),
    toolbar: (title) => ({
      type: 'NgToolbar',
      props: {
        title,
        backgroundColor: 'primary',
        textColor: 'white',
        class: 'q-mb-md',
        disableAlignEnd: true,
        rounded: true,
      },
      append,
      attrs,
      on,
      children: {
        default: () => {
          return [
            // child nodes or string content
          ]
        }
      },
      createMenu(children = () => []) {
        this.children.default = children
        return this
      }
    }),
    card: (title, subtitle) => ({
      type: 'NgCard',
      props: {
        title,
        subtitle,
        disableToolbar: true,
      },
      children: {
        default: () => [],
      },
      append,
      attrs,
      on,
    }),
    section: (title, subtitle) => ({
      type: 'NgCardSection',
      props: {
        title,
        subtitle,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    sectionAction: (align = 'right') => ({
      type: 'NgCardActions',
      props: {
        align,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    sectionRow: (title, subtitle, justify = 'start') => ({
      type: 'NgCardSection',
      props: {
        title,
        subtitle,
        class: `row justify-${justify} q-gutter-sm no-wrap`,
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,
    }),
    hline: (inset) => ({
      type: 'NgSeparator',
      props: {
        inset,
      },
      append,
      attrs,
      on,
    }),
    vline: (inset) => ({
      type: 'NgSeparator',
      props: {
        inset,
        vertical: true,
      },
      append,
      attrs,
      on,
    }),
    loading: (label, value, onUpdate = () => {}) => ({
      type: 'NgLoading',
      props: {
        label,
        modelValue: value, // assign value from data object with this.[variableName]
        'onUpdate:modelValue': onUpdate, // assign value from data object with update value
      },
      children: {
        default: () => []
      },
      append,
      attrs,
      on,   
    }),
    bannerNegative: (text) => ({
      type: 'NgBanner',
      props: {
        text,
        color: 'negative',
        icon: 'notifications',
        iconColor: 'white',
        textColor: 'white',
        rounded: true,
      },
      append,
      attrs,
      on,   
    }),
    bannerWarn: (text) => ({
      type: 'NgBanner',
      props: {
        text,
        color: 'warning',
        icon: 'warning',
        iconColor: 'white',
        textColor: 'white',
        rounded: true,
      },
      append,
      attrs,
      on,   
    }),
    bannerInfo: (text) => ({
      type: 'NgBanner',
      props: {
        text,
        color: 'grey-7',
        icon: 'info',
        iconColor: 'white',
        textColor: 'white',
        rounded: true,
      },
      append,
      attrs,
      on,   
    }),
    bannerPositive: (text) => ({
      type: 'NgBanner',
      props: {
        text,
        color: 'positive',
        icon: 'done',
        iconColor: 'white',
        textColor: 'white',
        rounded: true,
      },
      append,
      attrs,
      on,   
    }),
    markdown: (text) => ({
      type: 'NgMarkdown',
      props: {
        markdown: text,
      },
      append,
      attrs,
      on,    
    }),
    scroll: (height = '300px') => ({
      type: 'NgScroll',
      props: {
        style: {
          height,
        },
      },
      append,
      attrs,
      on,    
    }),
    chart: (title, opts = {}, series, maxHeight = 200) => {
      return {
        type: 'NgChart',
        props: {
          disableToolbars: [
            'left', 'right', 'reset', 'zoom', 'play', 'stop', 'settings', 'range', 'dates', 'select', 'download',
          ],
          disableTimebar: true,
          maxHeight,
          class: 'full-width',
          title,
          opts,
          series,
        },
        append,
        attrs,
        on,    
      }
    },
    processConfig: (configRef, processManagerUuid, instanceUuid, processUuid, config = undefined) => ({
      type: 'NgConfig',
      props: {
        configRef,
        processManagerUuid,
		instanceUuid,
        processUuid,
        config,
      },
      append,
      attrs,
      on,    
    }),
  }

  function getButtonsGrid(items = [{ roles: [], permissions: [], icon: '', color: '', to: '', href: '', disable: false, name: '', description: '' }], query= {}) {
    return context.grid(parent.ctx.platform.mobile ? 'start' : 'start', 'start', 'sm').append(() => [
      ...items.map((item) => window.components.element('NgSection').attrs({ 
          class: 'col-12 col-xs-6 col-md-4 col-lg-2', 
          permissions: item.permissions,
          roles: item.roles || [],
        }).append(() => [
          window.components.element('div').attrs({ class: 'full-width' }).append(() => [
            window.components.element('NgActionButton').attrs({
              color: '',
              fullWidth: true,
              disable: item.disable,
              padding: 'none',
              push: false,
              square: false,
              onClick: () =>  {
                if (item.to) {
                  context.router().push(item.to, item.query ? {...query, ...item.query} : query)
                } else {
                  parent.ctx.routing.openURL(item.href)
                }
              },
            }).append(() => [
              window.components.card().attrs({ class: 'full-width' }).append(() => [
                window.components.section().attrs({ class: `bg-${item.color} q-pa-none` }).append(() => [
                  window.components.section().attrs({ class: `flex justify-center items-center` }).append(() => [
                    window.components.element('NgIcon').attrs({ name: item.icon, size: '2rem', color: 'white' }),
                    window.components.renderIf(item.href, window.components.element('NgIcon').attrs({ name: 'mdi-open-in-new', size: '0.85rem', color: 'white' })),
                  ]),
                ]),
                window.components.section().attrs({ class: 'q-pa-sm', style: { background: parent.ctx.colors.lighten(parent.ctx.colors.getPaletteColor(item.color), parent.ctx.application.ctx.isDark ? -20 : 90) } }).append(() => [
                  window.components.text(item.name).attrs({ type: '' }),
                  window.components.text(item.description).attrs({ type: 'subtitle2' }),     
                ]),
              ])
            ])
          ])
        ])),
    ])
  }

  return {
    ...context,
    getButtonsGrid,
  }
}