// Call ngui function to create a ui config.
const FormInstance = window.nguiAsyncImport('public/forms/instance.js')
const FormWorkspace = window.nguiAsyncImport('public/forms/workspace.js')

window.ngui(() => {
  return {
    mixins: [
      window.processManager.mixin()
    ],
    beforeCreate() {
      window.components.setParent(this)
    }, 

    // The returned object can be accessed with this within any method...
    // Do not use this.ctx within this function because it is not yet created.
    data() {
      return {
        // VAR
        editWorkspaceUuid: '',
		exportWorkspaceUuid: '',
        dialog: {
          remove: false,
          add: false,
          addProject: false,
		  exportWorkspace: false,
		  importWorkspace: false,
        },
        apps: [],
        selectable: false,
        loading: false,
        managers: [],
        managerTabItems: [],
        activeProcessManagerUuid: '',
        workspacesByProcessManager: {} //  like {"processManagerUuid": []}
      }
    },

    methods: {  
      async init() {
        this.loading = true
        this.managers = await this.ctx.processManager.fetchManagers()
        this.createManagerTabs(this.managers)
        this.loading = false
      },

      async fetchWorkspaces(processManagerUuid) {
        this.loading = true
        try {
          const workspaces = await this.ctx.workspace.toService([processManagerUuid], true).fetchWorkspaces()
          this.workspacesByProcessManager[processManagerUuid] = workspaces
        } catch (e) {
          this.ctx.notify.notifyApiError(e)
        }

        this.loading = false
      },

      createManagerTabs(managers = []) {
        this.managerTabItems = managers.map((manager) => ({
          name: manager.options.slug,
          label: manager.options.name,
        }))

        const prevTab = this.ctx.rest.lsGetJSON('workspace:prev:managerTab')
        const hasPrevTab = !!this.managerTabItems.find((item) => item.name === prevTab)
        if (hasPrevTab) {
          this.activeProcessManagerUuid = prevTab
        } else {
          this.activeProcessManagerUuid = 'local-process-manager'
        }

        this.fetchWorkspaces(this.activeProcessManagerUuid)
      },
      
      removeAllSelected() {
        this.removeSelectedWorkspaces(this.activeProcessManagerUuid).then(() => this.init())
      },

      handleClickMenu(workspace, action) {
        switch (action) {
          case 'edit':
            this.editWorkspaceUuid = workspace.uuid
            this.dialog.add = true
            break
          case 'delete':
            this.selectedWorkspaces = [workspace]
            this.dialog.remove = true
            break
		  case 'export':
            this.exportWorkspaceUuid = workspace.uuid
            this.dialog.exportWorkspace = true
            break
        }
      },

      getAmount(workspace) {
        const nts = workspace.promosnts.length
        const instances = workspace.instances.length
        return nts + instances
      }
    },
      
    // Define elements within the ui configuration...
    nodes() {
      return [
        // any ui component node can be used...
        window.components.page(this.ctx.globals.getPermissions(['workspacesRead']), () => this.init()).append(() => [
          window.components.breadcrumbs([
            {
              label: this.ctx.i18n.i18n.t('workspaces'),
              icon: 'fas fa-layer-group',
              tooltip: 'workspaces',
            }
          ]),
          window.components.bar('workspaces').createMenu('file', [
            {
              id: 'addProject',
              label: 'newProject',
              icon: 'add',
              separator: true,
              permissions: this.ctx.globals.getPermissions(['workspacesCreate']),
              click: () => {
                this.dialog.addProject = true
              }
            },
			{
              id: 'importWorkspace',
              label: 'importWorkspace',
              icon: 'upload',
              separator: true,
              permissions: this.ctx.globals.getPermissions(['workspacesCreate', 'workspacesUpdate']),
              click: () => {
                this.dialog.importWorkspace = true
              }
            },
          ]).attrs({
            id: 'workspaces-toolbar',
          }),
          
          window.components.section().append(() => [

            window.components.grid().append(() => [
              window.components.cell(12).append(() => [
                window.components.element('NgTabs').attrs({
                  modelValue: this.activeProcessManagerUuid,
                  'onUpdate:modelValue': (val) => {
                    this.fetchWorkspaces(val)
                    this.activeProcessManagerUuid = val
                    this.ctx.rest.lsSetJSON('workspace:prev:managerTab', this.activeProcessManagerUuid)
                    this.selectedWorkspaces = []
                  },
                  tabs: this.managerTabItems,
                  disableTitle: true,
                  disableQueryTab: true,
                }).append(this.activeProcessManagerUuid, () => [
                   // Workspaces Table
                    window.components.card('workspaces').attrs({ flat: true }).append(() => [
                      window.components.table(
                        '', 
                        [
                          {
                            i18nLabel: 'name',
                            align: 'left',
                            name: 'name',
                            field: 'name',
                            sortable: true,
                          },
                          {
                            i18nLabel: 'description',
                            align: 'left',
                            name: 'description',
                            field: 'description',
                            sortable: true,
                          },
                          {
                            i18nLabel: 'projects',
                            align: 'left',
                            name: 'projects',
                            field: 'projects',
                            sortable: true,
                          },
                          {
                            i18nLabel: 'action',
                            align: 'left',
                            name: 'action',
                            field: 'action'
                          }
                        ], 
                        this.workspacesByProcessManager[this.activeProcessManagerUuid] || [],
                      ).attrs({
                        id: 'workspaces-table',
                        selection: 'multiple',
                        selected: this.selectedWorkspaces,
                        rowKey: 'uuid',
                        grid: !this.ctx.platform.browser,
                        flat: true,
                        loading: this.loading,
                        'onUpdate:selected': (val) => this.selectedWorkspaces = val, 
                        'onClick:row': (props) => window.components.router().push('/workspaces/index', { 
                          service: this.activeProcessManagerUuid,
                          workspace: props.row.uuid, 
                        }),
                        enableAddBtn: true,
                        enableDeleteBtn: true,
                        permissionsAdd: this.ctx.globals.getPermissions(['workspacesCreate']),
                        permissionsDelete: this.ctx.globals.getPermissions(['workspacesDelete']),
                        permissionsSelect: this.ctx.globals.getPermissions(['workspacesDelete']),
                        'onClick:add': () => {
                          this.editWorkspaceUuid = ''
                          this.dialog.add = true
                        },
                        'onConfirm:delete': () => {
                          this.removeAllSelected()
                        },
                        maxHeight: this.ctx.platform.browser ? '69vh' : undefined,
                      }).append('body-cell-projects', (props) => [
                        window.components.td().append(() => [
                            this.getAmount(props.row)
                        ])
                      ]).append('body-cell-action', (props) => [
                        window.components.td().on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
                          window.components.workspaceMenu('', 'more_vert', {
                            update: this.ctx.globals.getPermissions(['workspacesUpdate']),
                            delete: this.ctx.globals.getPermissions(['workspacesDelete']),
							read: this.ctx.globals.getPermissions(['workspacesRead']),
                          }).on({
                            'onClick:menu': (action) => this.handleClickMenu(props.row, action),
                            flat: true,
                            dense: true,
                            round: true,
                          }),
                        ])
                      ]).append('item', (props) => [
                        window.components.tdItem(window.components.card().append(() => [
                          window.components.renderIf(this.selectable, window.components.section().append(() => [
                            window.components.input('select', 'checkbox', props.selected, (val) => props.selected = val).attrs({
                              permissions: this.ctx.globals.getPermissions(['workspacesDelete'])
                            }),
                          ])),
                          window.components.section().append(() => [
                            window.components.item().append('section', () => [
                              window.components.buttonFlatLeft(props.row.name).on({ onClick: () => window.components.router().push('/workspaces/index', { 
                                workspace: props.row.uuid,
                                service: this.activeProcessManagerUuid,
                              })}),
                              window.components.caption(props.row.description)
                            ]).append('side', () => [
                              window.components.element('div').on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
                                window.components.defaultMenu('', 'more_vert', {
                                  update: this.ctx.globals.getPermissions(['workspacesUpdate']),
                                  delete: this.ctx.globals.getPermissions(['workspacesDelete'])
                                }).on({
                                  'onClick:menu': (action) => this.handleClickMenu(props.row, action),
                                  flat: true,
                                  dense: true,
                                  round: true,
                                }),
                              ])
                            ]),
                            window.components.item('projects', this.getAmount(props.row)),
                          ]),
                        ]))
                      ])
                    ]),
                ]),
            
              ]),
            ]),

            window.components.dialogDelete(this.selectedWorkspaces.map((item) => item.name).join(', '), this.dialog.remove, (val) => this.dialog.remove = val).on({
              onOk: () => this.removeAllSelected(),
              onCancel: () => this.selectedWorkspaces = [],
            }),  

            window.components.dialog().attrs({
              title: 'workspace',
              seamless: false,
              modelValue: this.dialog.add,
              'onUpdate:modelValue': (val) => this.dialog.add = val,
            }).append(() => [
              window.components.element(FormWorkspace).attrs({
                workspaceUuid: this.editWorkspaceUuid,
                serviceUuid: this.activeProcessManagerUuid,
                onCancel: () => this.dialog.add = false,
                onSubmit: () => {
                  this.dialog.add = false
                  this.init()
                },
              })
            ]),
			
			window.components.dialog().attrs({
              title: 'exportWorkspace',
              seamless: false,
              modelValue: this.dialog.exportWorkspace,
              'onUpdate:modelValue': (val) => this.dialog.exportWorkspace = val,
            }).append(() => [
				window.components.element('NgWorkspaceExport').attrs({
                  workspaceUuid: this.exportWorkspaceUuid,
				  serviceUuid: this.activeProcessManagerUuid,
                })
            ]),
			
			window.components.dialog().attrs({
              title: 'importWorkspace',
              seamless: false,
              modelValue: this.dialog.importWorkspace,
              'onUpdate:modelValue': (val) => this.dialog.importWorkspace = val,
            }).append(() => [
				window.components.element('NgWorkspaceImport').attrs({
				  serviceUuid: this.activeProcessManagerUuid,
                })
            ]),
            
            window.components.dialog().attrs({
              title: 'project',
              seamless: false,
              modelValue: this.dialog.addProject,
              'onUpdate:modelValue': (val) => this.dialog.addProject = val,
            }).append(() => [
              window.components.element(FormInstance).attrs({
                withDMS: true,
                instanceUuid: '',
                serviceUuid: this.activeProcessManagerUuid,
                onCancel: () => this.dialog.addProject = false,
                onSubmit: (instance, form) => {
                  this.dialog.addProject = false
                  this.ctx.workspace.postWorkspace({
                    name: form.name,
                    description: form.description,
                    uuid: this.ctx.uuidv4(),
                    promosnts: [],
                    instances: [
                      {
                        instanceUuid: form.uuid,
                        processManagerUuid: 'local-process-manager',
                        uuid: this.ctx.uuidv4(),
                      }
                    ]
                  })
                  .then(() => {
                    this.init()
                  })
                }
              })
            ]),
          ]),
        ])
      ]
    },
  }
})