// Copyright 2022 MST Systemtechnik AG. All Rights Reserved.

window.addEventListener('load', function () {

    'use strict';

    // Error panel
    let modalTimer = 0;
    function showError(title, text, displayDuration) {
        let infoPanelEl = document.getElementById('error-panel');
        infoPanelEl.innerHTML = '<b>' + title + '</b><br>';
        infoPanelEl.innerHTML += text;
        infoPanelEl.className = 'sticky';
        // Start timer for auto close
        if (modalTimer) {
            clearTimeout(modalTimer);
        }
        if (displayDuration) {
            modalTimer = setTimeout(function () {
                infoPanelEl.className = 'sticky invisible';
            }, displayDuration);
        }
    }
    
    // Status request handling
    let statusRequest = null;
    let statusTimer = 0;

    function getXmlHttpStatusRequestObject() {
        if (!statusRequest) {
            // Create a new XMLHttpRequest object
            statusRequest = new XMLHttpRequest();
        }
        return statusRequest;
    }

    function readVmInfo() {
        let infoRequest = new XMLHttpRequest();
        let url = '/v1/top/status';
        infoRequest.onreadystatechange = function () {
            // Check response is ready or not
            if (infoRequest.readyState === 4) {
                if (infoRequest.status === 200) {
                    // Update info
                    let computerName = JSON.parse(infoRequest.responseText).mainInfo.computerName;
                    let topLinkEl = document.getElementById('top-link');
                    topLinkEl.innerHTML = '<a href="/top.html" target="_blank"><b>Top</b> [' + computerName + ']</a>';
                } 
            }
        };
        // asynchronous requests
        infoRequest.open("GET", url, true);
        infoRequest.timeout = 5000;
        // Send the request over the network
        infoRequest.send();
    }

    function pollInstancesStatus() {
        if (statusTimer) {
            clearTimeout(statusTimer);
            statusTimer = 0;
        }
        let now = new Date();
        statusRequest = getXmlHttpStatusRequestObject();
        let url = '/v1/instances/status?_' + now.getTime();
        statusRequest.onreadystatechange = function () {
            // Check response is ready or not
            if (statusRequest.readyState === 4) {
                if (statusRequest.status === 200) {
                    // Create / update status list
                    updateStatusList(JSON.parse(statusRequest.responseText));
                    // Update the live data every 1 second
                    statusTimer = setTimeout(pollInstancesStatus, 1000);
                } else {
                    let statusText = statusRequest.status + ' - ' + statusRequest.statusText;
                    if (statusRequest.status === 0) {
                        statusText = 'No Connection';
                    }
                    showError('Status Poll', 'Error ' + statusText + '<br><br>' + statusRequest.responseText, 6000);
                    // Retry in 5 sec
                    statusTimer = setTimeout(pollInstancesStatus, 5000);
                }
            }
        };
        // asynchronous requests
        statusRequest.open("GET", url, true);
        statusRequest.timeout = 4000;
        // Send the request over the network
        statusRequest.send();
    }

    // Action functions
    function doAction(instanceUuid, processUuid, action) {
        let now = new Date();
        let url = '/v1/instances/actions/' + instanceUuid;
        if (processUuid !== '') {
            url += '/processes/' + processUuid;
        }
        url += '?action=' + action;
        url += '&_' + now.getTime();
        let actionRequest = new XMLHttpRequest();
        actionRequest.onreadystatechange = function () {
            if (actionRequest.readyState === 2) {
                // showError('Action ' + action, 'Sent', 2000);
            } else if (actionRequest.readyState === 4) {
                if (actionRequest.status === 200) {
                    // showError('Action ' + action, 'Ok', 1000);
                    if (action === 'Rescan') {
                        location.reload();
                    }
                    else {
                        // Get new status in 50ms
                        if (statusTimer) {
                            clearTimeout(statusTimer);
                            statusTimer = 0;
                        }
                        statusTimer = setTimeout(pollInstancesStatus, 50);
                    }
                } else {
                    showError('Action ' + action, 'failed, code ' + actionRequest.status + ' - ' + actionRequest.statusText + '<br><br>' + actionRequest.responseText, 5000);
                }
            }
        };
        actionRequest.open("PUT", url, true);
        actionRequest.timeout = 4000;
        actionRequest.send();
    }

    function addActionButton(instanceUuid, processUuid, text, action, classSuffix) {
        let btn = document.createElement('button');
        btn.innerHTML = text;
        if (instanceUuid === '') {
            btn.className = 'btn-action btn-' + classSuffix;
        }
        else {
            btn.className = 'btn-action btn-invisible btn-' + classSuffix;
        }
        btn.addEventListener('click', function () {
            if (instanceUuid === '' && (action === 'Stop' || action === 'Restart')) {
                if (confirm('Are you sure to ' + classSuffix + ' all instances ?')) {
                    doAction(instanceUuid, processUuid, action);
                }
            }
            else if (action === 'Rescan') {
                if (confirm('Rescan will synchronise all instances to existing project directories (add new / delete no more existing)!')) {
                    doAction(instanceUuid, processUuid, action);
                }
            }
            else {
                doAction(instanceUuid, processUuid, action);
            }
        });
        return btn;
    }

    // Add main buttons
    let scanActionsEl = document.getElementById('rescan-action');
    scanActionsEl.appendChild(addActionButton('', '', 'Rescan instances', 'Rescan', 'rescan'));

    let mainActionsEl = document.getElementById('main-actions');
    mainActionsEl.appendChild(addActionButton('', '', 'Restart', 'Restart', 'restart'));
    mainActionsEl.appendChild(addActionButton('', '', 'Stop', 'Stop', 'stop'));
    mainActionsEl.appendChild(addActionButton('', '', 'Start', 'Start', 'start'));
    
    // Read vm info once
    readVmInfo();
    // Start first poll
    pollInstancesStatus();

    function addSwitchButton(instanceUuid, processUuid, text, actionOn, actionOff, classSuffix) {
        let btnSwitch = document.createElement('label');
        btnSwitch.innerHTML = text;
        btnSwitch.className = 'btn-switch btn-switch-' + classSuffix;
        let inputEl = document.createElement('input');
        inputEl.type = 'checkbox';
        inputEl.checked = false;
        btnSwitch.appendChild(inputEl);
        btnSwitch.appendChild(document.createElement('span'));
        btnSwitch.addEventListener('change', function () {
            if (inputEl.checked) {
                doAction(instanceUuid, processUuid, actionOn);
            }
            else {
                doAction(instanceUuid, processUuid, actionOff);
            }
        });
        return btnSwitch;
    }

    function createSpanElementWithClass(className) {
        let element = document.createElement('span');
        element.className = className;
        return element;
    }

    function secToShortString(sec) {
        if (sec > 86400) {
            return Number.parseFloat(sec / 86400).toFixed(2) + 'd';
        } else if (sec > 3600) {
            return Number.parseFloat(sec / 3600).toFixed(2) + 'h';
        } else if (sec > 60) {
            return Number.parseFloat(sec / 60).toFixed(1) + 'm';
        } else {
            return sec + 's';
        }
    }

    function updateStatusList(instanceArr) {

        let totalCpu = 0;
        let totalMem = 0;

        // Iterate all instances
        instanceArr.forEach(function (instanceObj) {
            // Find instance
            let instanceEl = document.getElementById(instanceObj.uuid);
            if (!instanceEl) {

                // Not yet created
                instanceEl = document.createElement('details');
                instanceEl.id = instanceObj.uuid;
                let summary = document.createElement('summary');

                // Add name
                let name = document.createElement('span');
                name.className = 'instance-name';
                // Add description and instance path as tooltip
                name.title = instanceObj.description + 
                    ' (Ports ' + instanceObj.portOffsetPrefix + '00-' + instanceObj.portOffsetPrefix + '99)';
                summary.appendChild(name);

                // Add buttons
                summary.appendChild(addActionButton(instanceObj.uuid, '', 'Restart', 'Restart', 'restart'));
                summary.appendChild(addActionButton(instanceObj.uuid, '', 'Stop', 'Stop', 'stop'));
                summary.appendChild(addActionButton(instanceObj.uuid, '', 'Start', 'Start', 'start'));
                summary.appendChild(addSwitchButton(instanceObj.uuid, '', '', 'Enable', 'Disable', 'instance-enable'));
                instanceEl.appendChild(summary);

                // Add status and info
                summary.appendChild(createSpanElementWithClass('instance-status'));
                summary.appendChild(createSpanElementWithClass('instance-info'));

                // Add process table
                let processTbl = document.createElement('table');
                let processTblHdr = document.createElement('thead');
                let processTableHdrStr = ['Name', 'PID', 'LastStart', 'LastStop', 'LastErrorText', 'Up-/CPU-Time', 'CPU %', 'Mem MB', 'Status', 'Enabled', 'Log', 'Sta', 'Actions'];
                processTableHdrStr.forEach(function (title) {
                    processTblHdr.appendChild(document.createElement("th")).appendChild(document.createTextNode(title));
                });
                processTbl.appendChild(processTblHdr);
                instanceEl.appendChild(processTbl);

                // Append to formular
                let instances = document.getElementById('instances');
                instances.appendChild(instanceEl);
            }

            // Update instance values
            instanceEl.querySelector('.instance-name').innerHTML = instanceObj.name;
            let instanceStatusEl = instanceEl.querySelector('.instance-status');
            if (instanceObj.statusText === 'Starting' || instanceObj.statusText === 'Terminating') {
                if (instanceObj.startTermProgress < 0) {
                    instanceStatusEl.innerHTML = instanceObj.statusText + ' (wait order ' + instanceObj.startOrder + ')';
                }
                else {
                    instanceStatusEl.innerHTML = instanceObj.statusText + ' (' + instanceObj.startTermProgress + '%)';
                }
            }
            else {
                instanceStatusEl.innerHTML = instanceObj.statusText;
            }
            instanceStatusEl.className = 'instance-status status status-' + instanceObj.statusText + (instanceObj.enabled ? '' : ' status-Disabled');
            instanceEl.querySelector('.btn-switch-instance-enable').querySelector('input').checked = instanceObj.enabled;

            if (instanceObj.statusText === 'Running' || instanceObj.statusText === 'PartialRunning' || instanceObj.name === '_ProcessManager') {
                // (partial) running
                let upTimeStr = '';
                if (instanceObj.name === '_ProcessManager') {
                    upTimeStr = 'Up-Time: <b>' + secToShortString(instanceObj.upTime) + '</b>, '
                }
                instanceEl.querySelector('.instance-info').innerHTML = upTimeStr +
                    'CPU-Time: <b>' + secToShortString(instanceObj.cpuTime) + '</b>, ' +
                    'CPU: <b>' + Number.parseFloat(instanceObj.cpuUsage).toFixed(1) + ' %</b>, ' + 
                    'Mem: <b>' + Number.parseFloat(instanceObj.memUsage / 1024 / 1024).toFixed(0) + ' MB</b>';
            }
            else {
                instanceEl.querySelector('.instance-info').innerHTML = '';
            }
            totalCpu += instanceObj.cpuUsage;
            totalMem += instanceObj.memUsage / 1024 / 1024;

            if (instanceObj.statusText === 'Starting' || instanceObj.statusText === 'Terminating') {
                instanceEl.querySelector('.btn-start').classList.add('btn-invisible');
                instanceEl.querySelector('.btn-stop').classList.add('btn-invisible');
                instanceEl.querySelector('.btn-restart').classList.add('btn-invisible');
            }
            else if (instanceObj.statusText === 'Running') {
                // Starting or (partial) running
                instanceEl.querySelector('.btn-start').classList.add('btn-invisible');
                instanceEl.querySelector('.btn-stop').classList.remove('btn-invisible');
                instanceEl.querySelector('.btn-restart').classList.remove('btn-invisible');
            } else {
                instanceEl.querySelector('.btn-start').classList.remove('btn-invisible');
                instanceEl.querySelector('.btn-stop').classList.remove('btn-invisible');
                instanceEl.querySelector('.btn-restart').classList.remove('btn-invisible');
            }

            // Iterate all processes
            instanceObj.processes.forEach(function (processObj) {
                // Find process
                let processTblRow = document.getElementById(processObj.uuid);
                if (!processTblRow) {

                    // Not yet created
                    processTblRow = instanceEl.querySelector('table').insertRow();
                    processTblRow.id = processObj.uuid;

                    // Add description
                    let description = document.createElement('span');
                    description.className = 'process-description';
                    // Add name and uuid as tooltip
                    description.title = processObj.name + '\n' + processObj.uuid;
                    processTblRow.insertCell().appendChild(description);

                    // Add other cells
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-pid'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-tsLastStart'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-tsLastStop'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-lastErrorText'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-timeinfo'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-cpuUsage'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-memUsage'));
                    processTblRow.insertCell().appendChild(createSpanElementWithClass('process-status'));
                    processTblRow.insertCell().appendChild(addSwitchButton(instanceObj.uuid, processObj.uuid, '', 'Enable', 'Disable', 'process-enable'));
                    if (processObj.uuid === "ProcessManager") {
                        processTblRow.querySelector('.btn-switch-process-enable').hidden = true;
                    }

                    // Add log links
                    let processTblRowLogLinks = processTblRow.insertCell();
                    let logLink = document.createElement('a');
                    let logLinkText = document.createTextNode('Log');
                    logLink.appendChild(logLinkText);
                    logLink.title = 'Log';
                    logLink.href = '/logs.html?processUuid=' + processObj.uuid;
                    logLink.target = '_blank';
                    processTblRowLogLinks.appendChild(logLink);

                    // Add status links
                    let processTblRowStatusLinks = processTblRow.insertCell();
                    let statusLink = document.createElement('a');
                    statusLink.className = 'status-link';
                    if (processObj.uuid === 'ProcessManager') {
                        let jobLink = document.createElement('a');
                        let jobLinkText = document.createTextNode('J');
                        jobLink.appendChild(jobLinkText);
                        jobLink.title = 'Running Jobs';
                        jobLink.href = '/jobs.html';
                        jobLink.target = '_blank';
                        processTblRowStatusLinks.appendChild(jobLink);
                        // Short status
                        statusLink.href = '/status.html?processUuid=ProcessManager&serverPort=' + processObj.statusServerPort;
                        let statusLinkText = document.createTextNode('S');
                        statusLink.appendChild(statusLinkText);
                        statusLink.title = 'Status';
                    }
                    else {
                        let statusLinkText = document.createTextNode('');
                        statusLink.appendChild(statusLinkText);
                        statusLink.title = 'No Status';
                    }
                    statusLink.target = '_blank';
                    processTblRowStatusLinks.appendChild(statusLink);

                    // Add buttons
                    let processTblRowActions = processTblRow.insertCell();
                    processTblRowActions.appendChild(addActionButton(instanceObj.uuid, processObj.uuid, 'Restart', 'Restart', 'restart'));
                    processTblRowActions.appendChild(addActionButton(instanceObj.uuid, processObj.uuid, 'Stop', 'Stop', 'stop'));
                    processTblRowActions.appendChild(addActionButton(instanceObj.uuid, processObj.uuid, 'Start', 'Start', 'start'));
                    processTblRowActions.appendChild(addActionButton(instanceObj.uuid, processObj.uuid, 'Execute', 'Execute', 'execute'));

                    // Append to instance
                    instanceEl.querySelector('table').appendChild(processTblRow);
                }

                // Update status link
                if (processObj.uuid !== 'ProcessManager') {
                    let statusLink = processTblRow.querySelector('.status-link');
                    if (processObj.statusServerPort == 0) {
                        statusLink.text = '';
                        statusLink.href = '';
                        statusLink.title = 'No Status';
                    }
                    else {
                        statusLink.text = 'Sta';
                        statusLink.href = '/status.html?processUuid=' + processObj.uuid + '&serverPort=' + processObj.statusServerPort;
                        statusLink.title = 'Status';
                    }
                }

                // Update process values
                processTblRow.querySelector('.process-description').innerHTML = processObj.description;
                if (processObj.tsLastStart > 0) {
                    processTblRow.querySelector('.process-tsLastStart').innerHTML = new Date(processObj.tsLastStart).toLocaleString();
                } else {
                    processTblRow.querySelector('.process-tsLastStart').innerHTML = '-';
                }
                if (processObj.tsLastStop > 0) {
                    processTblRow.querySelector('.process-tsLastStop').innerHTML = new Date(processObj.tsLastStop).toLocaleString();
                } else {
                    processTblRow.querySelector('.process-tsLastStop').innerHTML = '-';
                }
                // Report only on running (or executing) state
                if (processObj.pid != 0) {
                    processTblRow.querySelector('.process-pid').innerHTML = processObj.pid;
                    processTblRow.querySelector('.process-timeinfo').innerHTML = secToShortString(processObj.upTime) + ' / ' + secToShortString(processObj.cpuTime);
                    processTblRow.querySelector('.process-cpuUsage').innerHTML = Number.parseFloat(processObj.cpuUsage).toFixed(1);
                    processTblRow.querySelector('.process-memUsage').innerHTML = Number.parseFloat(processObj.memUsage / 1024 / 1024).toFixed(2);
                } else {
                    processTblRow.querySelector('.process-pid').innerHTML = '-';
                    processTblRow.querySelector('.process-timeinfo').innerHTML = '-';
                    processTblRow.querySelector('.process-cpuUsage').innerHTML = '-';
                    processTblRow.querySelector('.process-memUsage').innerHTML = '-';
                }
                processTblRow.querySelector('.process-lastErrorText').innerHTML = processObj.lastErrorText;
                processTblRow.querySelector('.process-lastErrorText').title = new Date(processObj.tsLastError).toLocaleString();
                let processStatusEl = processTblRow.querySelector('.process-status');
                if (processObj.statusText === 'Starting' || processObj.statusText === 'Terminating' || processObj.statusText === 'Executing') {
                    if (processObj.startTermProgress < 0) {
                        if (processObj.tsNextStart > 0) {
                            let waitTime = (processObj.tsNextStart - Date.now()) / 1000;
                            if (waitTime < 0) {
                                waitTime = 0;
                            }
                            processStatusEl.innerHTML = processObj.statusText + ' (wait ' + Number.parseFloat(waitTime).toFixed(0) + 's)';
                        }
                        else {
                            processStatusEl.innerHTML = processObj.statusText + ' (wait order ' + processObj.startOrder + ')';
                        }
                    }
                    else {
                        processStatusEl.innerHTML = processObj.statusText + ' (' + processObj.startTermProgress + '%)';
                    }
                }
                else {
                    if (processObj.statusText === 'Scheduled') {
                        processStatusEl.innerHTML = processObj.statusText + ' ' + new Date(processObj.tsNextStart).toLocaleString();
                    }
                    else {
                        processStatusEl.innerHTML = processObj.statusText;
                    }
                }
                processStatusEl.className = 'process-status status status-' + processObj.statusText + (processObj.enabled ? '' : ' status-Disabled');

                processTblRow.querySelector('.btn-switch-process-enable').querySelector('input').checked = processObj.enabled;

                processTblRow.querySelector('.btn-execute').classList.add('btn-invisible');
                processTblRow.querySelector('.btn-start').classList.add('btn-invisible');
                processTblRow.querySelector('.btn-restart').classList.add('btn-invisible');
                processTblRow.querySelector('.btn-stop').classList.add('btn-invisible');

                if (processObj.statusText === 'Starting' || processObj.statusText === 'Terminating' || processObj.uuid === "ProcessManager") {
                    // Let all buttons hidden
                }
                else if (processObj.statusText === 'Scheduled' || processObj.statusText === 'RunOnce') {
                    processTblRow.querySelector('.btn-execute').classList.remove('btn-invisible');
                    processTblRow.querySelector('.btn-restart').classList.remove('btn-invisible');
                    processTblRow.querySelector('.btn-stop').classList.remove('btn-invisible');
                } 
                else if (processObj.pid != 0) {
                    processTblRow.querySelector('.btn-restart').classList.remove('btn-invisible');
                    processTblRow.querySelector('.btn-stop').classList.remove('btn-invisible');
                }
                else {
                    processTblRow.querySelector('.btn-start').classList.remove('btn-invisible');
                }
            });
        });

        let totalInfoEl = document.getElementById('total-info');
        totalInfoEl.innerHTML = 
            'Total CPU: <b>' + Number.parseFloat(totalCpu).toFixed(0) + ' %</b>,<br>' + 
            'Total Mem: <b>' + Number.parseFloat(totalMem).toFixed(0) + ' MB</b>';
    }

});